#lang racket/base
(require "runtime.ss" schemeunit
         racket/match
         (for-syntax racket/base 
                     racket/pretty
                     racket/dict
                     syntax/id-table))
(provide (except-out (all-from-out "runtime.ss") 
                     #%module-begin
                     ∧ ∨ ⊕)
         (rename-out [module-begin #%module-begin]
                     [u:∨ ∨]
                     [u:∧ ∧]
                     [u:⊕ ⊕]))

(define unknown '?)

(define u:∨
  (match-lambda* 
    [(list #t _) #t]
    [(list _ #t) #t]
    [(list #f #f) #f]
    [(list _ _) unknown])) 

(define u:∧ 
  (match-lambda* 
    [(list #f _) #f]
    [(list _ #f) #f]
    [(list #t #t) #t]
    [(list _ _) unknown]))

(define u:⊕
  (match-lambda*
    [(list #t #t) #f]
    [(list #f #f) #f]
    [(list #f #t) #t]
    [(list #t #f) #t]
    [(list _ _) unknown]))

;; check to make sure that the unknown-propogating versions
;; of the boolean functions are consistent with the real ones
(for ([u:f (list u:∧ u:∨ u:⊕)]
      [f (list ∧ ∨ ⊕)])
  (for ([b1/u (in-list (list #t #f unknown))])
    (for ([b2/u (in-list (list #t #f unknown))])
      (define u:ans (u:f b1/u b2/u))
      (when (boolean? u:ans)
        (for ([b1 (if (boolean? b1/u)
                      (list b1/u)
                      (list #t #f))])
          (for ([b2 (if (boolean? b2/u)
                        (list b2/u)
                        (list #t #f))])
            (define ans (f b1 b2))
            (check-equal? u:ans ans)))))))

(define (run-example all-possibilities? func)
  (cond
    [all-possibilities?
      (let ([ht (make-hash)])
        (time (let loop ([i (- (procedure-arity func) 1)]
                         [bools '()])
                (cond
                  [(zero? i)
                   (let ([depth (+ 1 (apply func void bools))]) ;; need to add one here to count the step that seeds the inputs
                     (hash-set! ht depth (cons bools (hash-ref ht depth '()))))]
                  [else 
                   (loop (- i 1) (cons #t bools))
                   (loop (- i 1) (cons #f bools))])))
        (let ([vals (sort (hash-map ht list) < #:key car)])
          (for-each
           (λ (val)
             (printf "~a inputs have gate delay ~a, e.g. ~s\n" 
                     (length (list-ref val 1))
                     (list-ref val 0)
                     (map (λ (x) (if x 1 0)) (car (list-ref val 1)))))
           vals)))]
    [else
     (void (func (λ (x y) (printf "~a ~a\n" x y))))]))

(define-syntax (module-begin stx)
  (syntax-case stx ()
    [(_ a ...) 
     (let ()
       (define vars '())
       (define input-vars '())
       (define pre-set-input-vars '())
       (define let-frames '())
       (define def-vars '())
       
       (define id-table (make-bound-id-table))
       
       (define all-possibilities? #f)
       
       (define (traverse-exp exp)
         (syntax-case (local-expand exp 'module-begin #f) (u:⊕ u:∧ u:∨)
           [(f e1 e2)
            (with-syntax ([x1 (traverse-exp #'e1)]
                          [x2 (traverse-exp #'e2)]
                          [(xres) (generate-temporaries (list #'f))])
              (set! let-frames (cons #`[(xres) (f x1 x2)] let-frames))
              (set! vars (cons #'xres vars))
              #'xres)]
           [x
            (identifier? #'x)
            (let ()
              (define the-id (dict-ref id-table #'x #f))
              (unless the-id
                (raise-syntax-error 'hdl "free variable" #'x))
              the-id)]
           [_ (raise-syntax-error 'hdl "unknown expression" exp)]))
       
       (for ([x (in-list (syntax->list #'(a ...)))])
         (let ([ans (local-expand x 'module-begin (list #'#%app #'inputs #'define-values))])
           (syntax-case ans (define-values inputs)
             [(inputs (these-vars ...))
              (let ()
                (define the-vars (syntax->list #'(these-vars ...)))
                (set! all-possibilities? #t)
                (for ([var (in-list the-vars)])
                  (dict-set! id-table var var))
                (set! input-vars (append (reverse the-vars) input-vars)))]
             [(inputs (these-vars ...) exp) 
              (let ()
                (unless (number? (syntax-e #'exp))
                  (raise-syntax-error 'hdl "expected a number as initializer" #'exp))
                (define the-vars (syntax->list #'(these-vars ...)))
                (set! pre-set-input-vars (append (reverse the-vars) pre-set-input-vars))
                (set! vars (append (reverse the-vars) vars))
                (for ([var (in-list the-vars)])
                  (dict-set! id-table var var))
                (set! let-frames
                      (cons (with-syntax ([(i ...) 
                                           (let ([count (length the-vars)])
                                             (build-list count (λ (x) (- count x 1))))])
                              #'[(these-vars ...) (let ([n exp]) (values (extract-bit i n) ...))])
                            let-frames)))]
             [(define-values (this-var) exp)
              (let () 
                (define final-id (traverse-exp #'exp))
                (dict-set! id-table #'this-var final-id)
                (set! vars (cons #'this-var vars))
                (set! def-vars (cons (list #'this-var final-id) def-vars)))]
             [exp
              (void)])))
       (with-syntax ([(vars ...) (reverse vars)]
                     [(input-vars ...) (reverse input-vars)]
                     [(pre-set-input-vars ...) (reverse pre-set-input-vars)])
         (with-syntax ([(old-vars ...) (generate-temporaries #'(vars ...))]
                       [(let-frames ...) (reverse let-frames)]
                       [((def-var int-var) ...) def-vars])
           #`(#%module-begin 
              (run-example
               #,all-possibilities?
               (λ (details input-vars ...)
                 (let loop ([i 0] [vars unknown] ...)
                   (let ([old-vars vars] ...)
                     (let-values (let-frames ...)
                       (cond [(and (eq? old-vars vars) ...)
                              i]
                             [else
                              (details i #,(if all-possibilities?
                                               #'#f
                                               #'(append 
                                                  (list 'pre-set-input-vars (show pre-set-input-vars)) ...
                                                  (list 'def-var (show int-var)) ...)))
                              (loop (+ i 1) vars ...)]))))))))))]))

(define (show x)
  (if (symbol? x) 
      x
      (if x
          1
          0)))